import pytest
from fastavro.schema import (
    FINGERPRINT_ALGORITHMS,
    fingerprint,
    to_parsing_canonical_form,
)


@pytest.mark.parametrize(
    "fingerprint", ["CRC-64-AVRO", "SHA-256", "MD5", "sha256", "md5"]
)
def test_required_fingerprints(fingerprint):
    assert fingerprint in FINGERPRINT_ALGORITHMS


def test_unknown_algorithm():
    unknown_algorithm = "UNKNOWN"
    assert unknown_algorithm not in FINGERPRINT_ALGORITHMS

    with pytest.raises(ValueError, match="Unknown schema fingerprint algorithm"):
        fingerprint("string", unknown_algorithm)


@pytest.mark.parametrize(
    "original_schema,algorithm,expected_fingerprint",
    [
        ("int", "CRC-64-AVRO", "8f5c393f1ad57572"),
        ("int", "md5", "ef524ea1b91e73173d938ade36c1db32"),
        (
            "int",
            "sha256",
            "3f2b87a9fe7cc9b13835598c3981cd45e3e355309e5090aa0933d7becb6fba45",
        ),
        (
            {"type": "int"},
            "CRC-64-AVRO",
            "8f5c393f1ad57572",
        ),
        (
            {"type": "int"},
            "md5",
            "ef524ea1b91e73173d938ade36c1db32",
        ),
        (
            {"type": "int"},
            "sha256",
            "3f2b87a9fe7cc9b13835598c3981cd45e3e355309e5090aa0933d7becb6fba45",
        ),
        (
            "float",
            "CRC-64-AVRO",
            "90d7a83ecb027c4d",
        ),
        (
            "float",
            "md5",
            "50a6b9db85da367a6d2df400a41758a6",
        ),
        (
            "float",
            "sha256",
            "1e71f9ec051d663f56b0d8e1fc84d71aa56ccfe9fa93aa20d10547a7abeb5cc0",
        ),
        (
            {"type": "float"},
            "CRC-64-AVRO",
            "90d7a83ecb027c4d",
        ),
        (
            {"type": "float"},
            "md5",
            "50a6b9db85da367a6d2df400a41758a6",
        ),
        (
            {"type": "float"},
            "sha256",
            "1e71f9ec051d663f56b0d8e1fc84d71aa56ccfe9fa93aa20d10547a7abeb5cc0",
        ),
        (
            "long",
            "CRC-64-AVRO",
            "b71df49344e154d0",
        ),
        (
            "long",
            "md5",
            "e1dd9a1ef98b451b53690370b393966b",
        ),
        (
            "long",
            "sha256",
            "c32c497df6730c97fa07362aa5023f37d49a027ec452360778114cf427965add",
        ),
        (
            {"type": "long"},
            "CRC-64-AVRO",
            "b71df49344e154d0",
        ),
        (
            {"type": "long"},
            "md5",
            "e1dd9a1ef98b451b53690370b393966b",
        ),
        (
            {"type": "long"},
            "sha256",
            "c32c497df6730c97fa07362aa5023f37d49a027ec452360778114cf427965add",
        ),
        (
            "double",
            "CRC-64-AVRO",
            "7e95ab32c035758e",
        ),
        (
            "double",
            "md5",
            "bfc71a62f38b99d6a93690deeb4b3af6",
        ),
        (
            "double",
            "sha256",
            "730a9a8c611681d7eef442e03c16c70d13bca3eb8b977bb403eaff52176af254",
        ),
        (
            {"type": "double"},
            "CRC-64-AVRO",
            "7e95ab32c035758e",
        ),
        (
            {"type": "double"},
            "md5",
            "bfc71a62f38b99d6a93690deeb4b3af6",
        ),
        (
            {"type": "double"},
            "sha256",
            "730a9a8c611681d7eef442e03c16c70d13bca3eb8b977bb403eaff52176af254",
        ),
        (
            "bytes",
            "CRC-64-AVRO",
            "651920c3da16c04f",
        ),
        (
            "bytes",
            "md5",
            "b462f06cb909be57c85008867784cde6",
        ),
        (
            "bytes",
            "sha256",
            "9ae507a9dd39ee5b7c7e285da2c0846521c8ae8d80feeae5504e0c981d53f5fa",
        ),
        (
            {"type": "bytes"},
            "CRC-64-AVRO",
            "651920c3da16c04f",
        ),
        (
            {"type": "bytes"},
            "md5",
            "b462f06cb909be57c85008867784cde6",
        ),
        (
            {"type": "bytes"},
            "sha256",
            "9ae507a9dd39ee5b7c7e285da2c0846521c8ae8d80feeae5504e0c981d53f5fa",
        ),
        (
            "string",
            "CRC-64-AVRO",
            "c70345637248018f",
        ),
        (
            "string",
            "md5",
            "095d71cf12556b9d5e330ad575b3df5d",
        ),
        (
            "string",
            "sha256",
            "e9e5c1c9e4f6277339d1bcde0733a59bd42f8731f449da6dc13010a916930d48",
        ),
        (
            {"type": "string"},
            "CRC-64-AVRO",
            "c70345637248018f",
        ),
        (
            {"type": "string"},
            "md5",
            "095d71cf12556b9d5e330ad575b3df5d",
        ),
        (
            {"type": "string"},
            "sha256",
            "e9e5c1c9e4f6277339d1bcde0733a59bd42f8731f449da6dc13010a916930d48",
        ),
        (
            "boolean",
            "CRC-64-AVRO",
            "64f7d4a478fc429f",
        ),
        (
            "boolean",
            "md5",
            "01f692b30d4a1c8a3e600b1440637f8f",
        ),
        (
            "boolean",
            "sha256",
            "a5b031ab62bc416d720c0410d802ea46b910c4fbe85c50a946ccc658b74e677e",
        ),
        (
            {"type": "boolean"},
            "CRC-64-AVRO",
            "64f7d4a478fc429f",
        ),
        (
            {"type": "boolean"},
            "md5",
            "01f692b30d4a1c8a3e600b1440637f8f",
        ),
        (
            {"type": "boolean"},
            "sha256",
            "a5b031ab62bc416d720c0410d802ea46b910c4fbe85c50a946ccc658b74e677e",
        ),
        (
            "null",
            "CRC-64-AVRO",
            "8a8f25cce724dd63",
        ),
        (
            "null",
            "md5",
            "9b41ef67651c18488a8b08bb67c75699",
        ),
        (
            "null",
            "sha256",
            "f072cbec3bf8841871d4284230c5e983dc211a56837aed862487148f947d1a1f",
        ),
        (
            {"type": "null"},
            "CRC-64-AVRO",
            "8a8f25cce724dd63",
        ),
        (
            {"type": "null"},
            "md5",
            "9b41ef67651c18488a8b08bb67c75699",
        ),
        (
            {"type": "null"},
            "sha256",
            "f072cbec3bf8841871d4284230c5e983dc211a56837aed862487148f947d1a1f",
        ),
        (
            {"type": "fixed", "name": "Test", "size": 1},
            "CRC-64-AVRO",
            "6869897b4049355b",
        ),
        (
            {"type": "fixed", "name": "Test", "size": 1},
            "md5",
            "db01bc515fcfcd2d4be82ed385288261",
        ),
        (
            {"type": "fixed", "name": "Test", "size": 1},
            "sha256",
            "f527116a6f44455697e935afc31dc60ad0f95caf35e1d9c9db62edb3ffeb9170",
        ),
        (
            {
                "type": "fixed",
                "name": "MyFixed",
                "namespace": "org.apache.hadoop.avro",
                "size": 1,
            },
            "CRC-64-AVRO",
            "fadbd138e85bdf45",
        ),
        (
            {
                "type": "fixed",
                "name": "MyFixed",
                "namespace": "org.apache.hadoop.avro",
                "size": 1,
            },
            "md5",
            "d74b3726484422711c465d49e857b1ba",
        ),
        (
            {
                "type": "fixed",
                "name": "MyFixed",
                "namespace": "org.apache.hadoop.avro",
                "size": 1,
            },
            "sha256",
            "28e493a44771cecc5deca4bd938cdc3d5a24cfe1f3760bc938fa1057df6334fc",
        ),
        (
            {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            "CRC-64-AVRO",
            "03a2f2c2e27f7a16",
        ),
        (
            {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            "md5",
            "d883f2a9b16ed085fcc5e4ca6c8f6ed1",
        ),
        (
            {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            "sha256",
            "9b51286144f87ce5aebdc61ca834379effa5a41ce6ac0938630ff246297caca8",
        ),
        (
            {"type": "array", "items": "long"},
            "CRC-64-AVRO",
            "715e2ea28bc91654",
        ),
        (
            {"type": "array", "items": "long"},
            "md5",
            "c1c387e8d6a58f0df749b698991b1f43",
        ),
        (
            {"type": "array", "items": "long"},
            "sha256",
            "f78e954167feb23dcb1ce01e8463cebf3408e0a4259e16f24bd38f6d0f1d578b",
        ),
        (
            {
                "type": "array",
                "items": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            "CRC-64-AVRO",
            "10d9ade1fa3a0387",
        ),
        (
            {
                "type": "array",
                "items": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            "md5",
            "cfc7b861c7cfef082a6ef082948893fa",
        ),
        (
            {
                "type": "array",
                "items": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            "sha256",
            "0d8edd49d7f7e9553668f133577bc99f842852b55d9f84f1f7511e4961aa685c",
        ),
        (
            {"type": "map", "values": "long"},
            "CRC-64-AVRO",
            "6f74f4e409b1334e",
        ),
        (
            {"type": "map", "values": "long"},
            "md5",
            "32b3f1a3177a0e73017920f00448b56e",
        ),
        (
            {"type": "map", "values": "long"},
            "sha256",
            "b8fad07d458971a07692206b8a7cf626c86c62fe6bcff7c1b11bc7295de34853",
        ),
        (
            {
                "type": "map",
                "values": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            "CRC-64-AVRO",
            "df2ab0626f6b812d",
        ),
        (
            {
                "type": "map",
                "values": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            "md5",
            "c588da6ba99701c41e73fd30d23f994e",
        ),
        (
            {
                "type": "map",
                "values": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            "sha256",
            "3886747ed1669a8af476b549e97b34222afb2fed5f18bb27c6f367ea0351a576",
        ),
        (
            ["string", "null", "long"],
            "CRC-64-AVRO",
            "65a5be410d687566",
        ),
        (
            ["string", "null", "long"],
            "md5",
            "b11cf95f0a55dd55f9ee515a37bf937a",
        ),
        (
            ["string", "null", "long"],
            "sha256",
            "ed8d254116441bb35e237ad0563cf5432b8c975334bd222c1ee84609435d95bb",
        ),
        (
            {
                "type": "record",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            "CRC-64-AVRO",
            "ed94e5f5e6eb588e",
        ),
        (
            {
                "type": "record",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            "md5",
            "69531a03db788afe353244cd049b1e6d",
        ),
        (
            {
                "type": "record",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            "sha256",
            "9670f15a8f96d23e92830d00b8bd57275e02e3e173ffef7c253c170b6beabeb8",
        ),
        (
            {
                "type": "error",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            "CRC-64-AVRO",
            "ed94e5f5e6eb588e",
        ),
        (
            {
                "type": "error",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            "md5",
            "69531a03db788afe353244cd049b1e6d",
        ),
        (
            {
                "type": "error",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            "sha256",
            "9670f15a8f96d23e92830d00b8bd57275e02e3e173ffef7c253c170b6beabeb8",
        ),
        (
            {
                "type": "record",
                "name": "Node",
                "fields": [
                    {"name": "label", "type": "string"},
                    {"name": "children", "type": {"type": "array", "items": "Node"}},
                ],
            },
            "CRC-64-AVRO",
            "52cba544c3e756b7",
        ),
        (
            {
                "type": "record",
                "name": "Node",
                "fields": [
                    {"name": "label", "type": "string"},
                    {"name": "children", "type": {"type": "array", "items": "Node"}},
                ],
            },
            "md5",
            "99625b0cc02050363e89ef66b0f406c9",
        ),
        (
            {
                "type": "record",
                "name": "Node",
                "fields": [
                    {"name": "label", "type": "string"},
                    {"name": "children", "type": {"type": "array", "items": "Node"}},
                ],
            },
            "sha256",
            "65d80dc8c95c98a9671d92cf0415edfabfee2cb058df2138606656cd6ae4dc59",
        ),
        (
            {
                "type": "record",
                "name": "Lisp",
                "fields": [
                    {
                        "name": "value",
                        "type": [
                            "null",
                            "string",
                            {
                                "type": "record",
                                "name": "Cons",
                                "fields": [
                                    {"name": "car", "type": "Lisp"},
                                    {"name": "cdr", "type": "Lisp"},
                                ],
                            },
                        ],
                    },
                ],
            },
            "CRC-64-AVRO",
            "68d91a23eda0b306",
        ),
        (
            {
                "type": "record",
                "name": "Lisp",
                "fields": [
                    {
                        "name": "value",
                        "type": [
                            "null",
                            "string",
                            {
                                "type": "record",
                                "name": "Cons",
                                "fields": [
                                    {"name": "car", "type": "Lisp"},
                                    {"name": "cdr", "type": "Lisp"},
                                ],
                            },
                        ],
                    },
                ],
            },
            "md5",
            "9e1d0d15b52789fcb8e3a88b53059d5f",
        ),
        (
            {
                "type": "record",
                "name": "Lisp",
                "fields": [
                    {
                        "name": "value",
                        "type": [
                            "null",
                            "string",
                            {
                                "type": "record",
                                "name": "Cons",
                                "fields": [
                                    {"name": "car", "type": "Lisp"},
                                    {"name": "cdr", "type": "Lisp"},
                                ],
                            },
                        ],
                    },
                ],
            },
            "sha256",
            "e5ce4f4a15ce19fa1047cfe16a3b0e13a755db40f00f23284fdd376fc1c7dd21",
        ),
        (
            {
                "type": "record",
                "name": "HandshakeRequest",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "clientHash",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {"name": "clientProtocol", "type": ["null", "string"]},
                    {"name": "serverHash", "type": "MD5"},
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            "CRC-64-AVRO",
            "b96ad79e5a7c5757",
        ),
        (
            {
                "type": "record",
                "name": "HandshakeRequest",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "clientHash",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {"name": "clientProtocol", "type": ["null", "string"]},
                    {"name": "serverHash", "type": "MD5"},
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            "md5",
            "4c822af2e17eecd92422827eede97f5b",
        ),
        (
            {
                "type": "record",
                "name": "HandshakeRequest",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "clientHash",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {"name": "clientProtocol", "type": ["null", "string"]},
                    {"name": "serverHash", "type": "MD5"},
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            "sha256",
            "2b2f7a9b22991fe0df9134cb6b5ff7355343e797aaea337e0150e20f3a35800e",
        ),
        (
            {
                "type": "record",
                "name": "HandshakeResponse",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "match",
                        "type": {
                            "type": "enum",
                            "name": "HandshakeMatch",
                            "symbols": ["BOTH", "CLIENT", "NONE"],
                        },
                    },
                    {"name": "serverProtocol", "type": ["null", "string"]},
                    {
                        "name": "serverHash",
                        "type": ["null", {"name": "MD5", "size": 16, "type": "fixed"}],
                    },
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            "CRC-64-AVRO",
            "00feee01de4ea50e",
        ),
        (
            {
                "type": "record",
                "name": "HandshakeResponse",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "match",
                        "type": {
                            "type": "enum",
                            "name": "HandshakeMatch",
                            "symbols": ["BOTH", "CLIENT", "NONE"],
                        },
                    },
                    {"name": "serverProtocol", "type": ["null", "string"]},
                    {
                        "name": "serverHash",
                        "type": ["null", {"name": "MD5", "size": 16, "type": "fixed"}],
                    },
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            "md5",
            "afe529d01132daab7f4e2a6663e7a2f5",
        ),
        (
            {
                "type": "record",
                "name": "HandshakeResponse",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "match",
                        "type": {
                            "type": "enum",
                            "name": "HandshakeMatch",
                            "symbols": ["BOTH", "CLIENT", "NONE"],
                        },
                    },
                    {"name": "serverProtocol", "type": ["null", "string"]},
                    {
                        "name": "serverHash",
                        "type": ["null", {"name": "MD5", "size": 16, "type": "fixed"}],
                    },
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            "sha256",
            "a303cbbfe13958f880605d70c521a4b7be34d9265ac5a848f25916a67b11d889",
        ),
        (
            {
                "type": "record",
                "name": "Interop",
                "namespace": "org.apache.avro",
                "fields": [
                    {"name": "intField", "type": "int"},
                    {"name": "longField", "type": "long"},
                    {"name": "stringField", "type": "string"},
                    {"name": "boolField", "type": "boolean"},
                    {"name": "floatField", "type": "float"},
                    {"name": "doubleField", "type": "double"},
                    {"name": "bytesField", "type": "bytes"},
                    {"name": "nullField", "type": "null"},
                    {
                        "name": "arrayField",
                        "type": {"type": "array", "items": "double"},
                    },
                    {
                        "name": "mapField",
                        "type": {
                            "type": "map",
                            "values": {
                                "name": "Foo",
                                "type": "record",
                                "fields": [{"name": "label", "type": "string"}],
                            },
                        },
                    },
                    {
                        "name": "unionField",
                        "type": [
                            "boolean",
                            "double",
                            {"type": "array", "items": "bytes"},
                        ],
                    },
                    {
                        "name": "enumField",
                        "type": {
                            "type": "enum",
                            "name": "Kind",
                            "symbols": ["A", "B", "C"],
                        },
                    },
                    {
                        "name": "fixedField",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {
                        "name": "recordField",
                        "type": {
                            "type": "record",
                            "name": "Node",
                            "fields": [
                                {"name": "label", "type": "string"},
                                {
                                    "name": "children",
                                    "type": {"type": "array", "items": "Node"},
                                },
                            ],
                        },
                    },
                ],
            },
            "CRC-64-AVRO",
            "e82c0a93a6a0b5a4",
        ),
        (
            {
                "type": "record",
                "name": "Interop",
                "namespace": "org.apache.avro",
                "fields": [
                    {"name": "intField", "type": "int"},
                    {"name": "longField", "type": "long"},
                    {"name": "stringField", "type": "string"},
                    {"name": "boolField", "type": "boolean"},
                    {"name": "floatField", "type": "float"},
                    {"name": "doubleField", "type": "double"},
                    {"name": "bytesField", "type": "bytes"},
                    {"name": "nullField", "type": "null"},
                    {
                        "name": "arrayField",
                        "type": {"type": "array", "items": "double"},
                    },
                    {
                        "name": "mapField",
                        "type": {
                            "type": "map",
                            "values": {
                                "name": "Foo",
                                "type": "record",
                                "fields": [{"name": "label", "type": "string"}],
                            },
                        },
                    },
                    {
                        "name": "unionField",
                        "type": [
                            "boolean",
                            "double",
                            {"type": "array", "items": "bytes"},
                        ],
                    },
                    {
                        "name": "enumField",
                        "type": {
                            "type": "enum",
                            "name": "Kind",
                            "symbols": ["A", "B", "C"],
                        },
                    },
                    {
                        "name": "fixedField",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {
                        "name": "recordField",
                        "type": {
                            "type": "record",
                            "name": "Node",
                            "fields": [
                                {"name": "label", "type": "string"},
                                {
                                    "name": "children",
                                    "type": {"type": "array", "items": "Node"},
                                },
                            ],
                        },
                    },
                ],
            },
            "md5",
            "994fea1a1be7ff8603cbe40c3bc7e4ca",
        ),
        (
            {
                "type": "record",
                "name": "Interop",
                "namespace": "org.apache.avro",
                "fields": [
                    {"name": "intField", "type": "int"},
                    {"name": "longField", "type": "long"},
                    {"name": "stringField", "type": "string"},
                    {"name": "boolField", "type": "boolean"},
                    {"name": "floatField", "type": "float"},
                    {"name": "doubleField", "type": "double"},
                    {"name": "bytesField", "type": "bytes"},
                    {"name": "nullField", "type": "null"},
                    {
                        "name": "arrayField",
                        "type": {"type": "array", "items": "double"},
                    },
                    {
                        "name": "mapField",
                        "type": {
                            "type": "map",
                            "values": {
                                "name": "Foo",
                                "type": "record",
                                "fields": [{"name": "label", "type": "string"}],
                            },
                        },
                    },
                    {
                        "name": "unionField",
                        "type": [
                            "boolean",
                            "double",
                            {"type": "array", "items": "bytes"},
                        ],
                    },
                    {
                        "name": "enumField",
                        "type": {
                            "type": "enum",
                            "name": "Kind",
                            "symbols": ["A", "B", "C"],
                        },
                    },
                    {
                        "name": "fixedField",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {
                        "name": "recordField",
                        "type": {
                            "type": "record",
                            "name": "Node",
                            "fields": [
                                {"name": "label", "type": "string"},
                                {
                                    "name": "children",
                                    "type": {"type": "array", "items": "Node"},
                                },
                            ],
                        },
                    },
                ],
            },
            "sha256",
            "cccfd6e3f917cf53b0f90c206342e6703b0d905071f724a1c1f85b731c74058d",
        ),
        (
            {
                "type": "record",
                "name": "ipAddr",
                "fields": [
                    {
                        "name": "addr",
                        "type": [
                            {"name": "IPv6", "type": "fixed", "size": 16},
                            {"name": "IPv4", "type": "fixed", "size": 4},
                        ],
                    }
                ],
            },
            "CRC-64-AVRO",
            "8d961b4e298a1844",
        ),
        (
            {
                "type": "record",
                "name": "ipAddr",
                "fields": [
                    {
                        "name": "addr",
                        "type": [
                            {"name": "IPv6", "type": "fixed", "size": 16},
                            {"name": "IPv4", "type": "fixed", "size": 4},
                        ],
                    }
                ],
            },
            "md5",
            "45d85c69b353a99b93d7c4f2fcf0c30d",
        ),
        (
            {
                "type": "record",
                "name": "ipAddr",
                "fields": [
                    {
                        "name": "addr",
                        "type": [
                            {"name": "IPv6", "type": "fixed", "size": 16},
                            {"name": "IPv4", "type": "fixed", "size": 4},
                        ],
                    }
                ],
            },
            "sha256",
            "6f6fc8f685a4f07d99734946565d63108806d55a8620febea047cf52cb0ac181",
        ),
        (
            {
                "type": "record",
                "name": "TestDoc",
                "doc": "Doc string",
                "fields": [{"name": "name", "type": "string", "doc": "Doc String"}],
            },
            "CRC-64-AVRO",
            "0e6660f02bcdc109",
        ),
        (
            {
                "type": "record",
                "name": "TestDoc",
                "doc": "Doc string",
                "fields": [{"name": "name", "type": "string", "doc": "Doc String"}],
            },
            "md5",
            "f2da75f5131f5ab80629538287b8beb2",
        ),
        (
            {
                "type": "record",
                "name": "TestDoc",
                "doc": "Doc string",
                "fields": [{"name": "name", "type": "string", "doc": "Doc String"}],
            },
            "sha256",
            "0b3644f7aa5ca2fc4bad93ca2d3609c12aa9dbda9c15e68b34c120beff08e7b9",
        ),
        (
            {
                "type": "enum",
                "name": "Test",
                "symbols": ["A", "B"],
                "doc": "Doc String",
            },
            "CRC-64-AVRO",
            "03a2f2c2e27f7a16",
        ),
        (
            {
                "type": "enum",
                "name": "Test",
                "symbols": ["A", "B"],
                "doc": "Doc String",
            },
            "md5",
            "d883f2a9b16ed085fcc5e4ca6c8f6ed1",
        ),
        (
            {
                "type": "enum",
                "name": "Test",
                "symbols": ["A", "B"],
                "doc": "Doc String",
            },
            "sha256",
            "9b51286144f87ce5aebdc61ca834379effa5a41ce6ac0938630ff246297caca8",
        ),
        (
            {"type": "int"},
            "MD5",  # JAVA Name
            "ef524ea1b91e73173d938ade36c1db32",
        ),
        (
            {"type": "int"},
            "SHA-256",  # JAVA Name
            "3f2b87a9fe7cc9b13835598c3981cd45e3e355309e5090aa0933d7becb6fba45",
        ),
    ],
)
def test_random_cases(original_schema, algorithm, expected_fingerprint):
    # All of these random test cases came from the test cases here:
    # https://github.com/apache/avro/blob/0552c674637dd15b8751ed5181387cdbd81480d5/lang/py3/avro/tests/test_normalization.py
    canonical_form = to_parsing_canonical_form(original_schema)
    assert fingerprint(canonical_form, algorithm) == expected_fingerprint



from typing import Callable, Dict, IO, Iterable, Optional, Union, Any
from .io.binary_encoder import BinaryEncoder
from .io.json_encoder import AvroJSONEncoder
from .types import AvroMessage, Schema

def writer(
    fo: Union[IO, AvroJSONEncoder],
    schema: Schema,
    records: Iterable[Any],
    codec: str = ...,
    sync_interval: int = ...,
    metadata: Optional[Dict[str, str]] = ...,
    validator: bool = ...,
    sync_marker: bytes = ...,
    codec_compression_level: Optional[int] = ...,
    *,
    strict: bool = ...,
    strict_allow_default: bool = ...,
    disable_tuple_notation: bool = ...,
) -> None: ...

class GenericWriter:
    schema: Schema
    validate_fn: Optional[Callable]
    metadata: Dict

class Writer(GenericWriter):
    encoder: BinaryEncoder
    io: BinaryEncoder
    block_count: int
    sync_interval: int
    compression_level: Optional[int]
    block_writer: Callable
    def __init__(
        self,
        fo: Union[IO, BinaryEncoder],
        schema: Schema,
        codec: str = ...,
        sync_interval: int = ...,
        metadata: Optional[Dict[str, str]] = ...,
        validator: bool = ...,
        sync_marker: bytes = ...,
        compression_level: Optional[int] = ...,
        options: Dict[str, bool] = ...,
    ): ...
    def dump(self) -> None: ...
    def write(self, record: AvroMessage) -> None: ...
    def write_block(self, block) -> None: ...  # type: ignore  # Should be a read.Block
    def flush(self) -> None: ...

def schemaless_writer(
    fo: IO,
    schema: Schema,
    record: Any,
    *,
    strict: bool = ...,
    strict_allow_default: bool = ...,
    disable_tuple_notation: bool = ...,
) -> None: ...



from typing import Any, Iterable

from .types import Schema

def validate(
    datum: Any,
    schema: Schema,
    field: str = ...,
    raise_errors: bool = ...,
    strict: bool = ...,
) -> bool: ...
def validate_many(
    records: Iterable[Any],
    schema: Schema,
    raise_errors: bool = ...,
    strict: bool = ...,
) -> bool: ...



# cython: language_level=3
# cython: auto_cpdef=True

"""Python code for reading AVRO files"""

# This code is a modified version of the code at
# http://svn.apache.org/viewvc/avro/trunk/lang/py/src/avro/ which is under
# Apache 2.0 license (http://www.apache.org/licenses/LICENSE-2.0)

import bz2
import lzma
import zlib
from datetime import datetime, timezone
from decimal import Context
from io import BytesIO

import json

from .logical_readers import LOGICAL_READERS
from ._schema import extract_record_type, is_nullable_union, extract_logical_type, parse_schema
from ._read_common import (
    SchemaResolutionError,
    MAGIC,
    SYNC_SIZE,
    HEADER_SCHEMA,
    missing_codec_lib,
)
from .const import NAMED_TYPES, AVRO_TYPES

CYTHON_MODULE = 1  # Tests check this to confirm whether using the Cython code.

decimal_context = Context()
epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
epoch_naive = datetime(1970, 1, 1)

ctypedef int int32
ctypedef unsigned int uint32
ctypedef unsigned long long ulong64
ctypedef long long long64


class ReadError(Exception):
    pass


cpdef match_types(writer_type, reader_type):
    if isinstance(writer_type, list) or isinstance(reader_type, list):
        return True
    if isinstance(writer_type, dict) or isinstance(reader_type, dict):
        try:
            return match_schemas(writer_type, reader_type)
        except SchemaResolutionError:
            return False
    if writer_type == reader_type:
        return True
    # promotion cases
    elif writer_type == "int" and reader_type in ["long", "float", "double"]:
        return True
    elif writer_type == "long" and reader_type in ["float", "double"]:
        return True
    elif writer_type == "float" and reader_type == "double":
        return True
    elif writer_type == "string" and reader_type == "bytes":
        return True
    elif writer_type == "bytes" and reader_type == "string":
        return True
    return False


cpdef match_schemas(w_schema, r_schema):
    error_msg = f"Schema mismatch: {w_schema} is not {r_schema}"
    if isinstance(w_schema, list):
        # If the writer is a union, checks will happen in read_union after the
        # correct schema is known
        return r_schema
    elif isinstance(r_schema, list):
        # If the reader is a union, ensure one of the new schemas is the same
        # as the writer
        for schema in r_schema:
            if match_types(w_schema, schema):
                return schema
        else:
            raise SchemaResolutionError(error_msg)
    else:
        # Check for dicts as primitive types are just strings
        if isinstance(w_schema, dict):
            w_type = w_schema["type"]
        else:
            w_type = w_schema
        if isinstance(r_schema, dict):
            r_type = r_schema["type"]
        else:
            r_type = r_schema

        if w_type == r_type == "map":
            if match_types(w_schema["values"], r_schema["values"]):
                return r_schema
        elif w_type == r_type == "array":
            if match_types(w_schema["items"], r_schema["items"]):
                return r_schema
        elif w_type in NAMED_TYPES and r_type in NAMED_TYPES:
            if w_type == r_type == "fixed" and w_schema["size"] != r_schema["size"]:
                raise SchemaResolutionError(
                    f"Schema mismatch: {w_schema} size is different than {r_schema} size"
                )

            w_unqual_name = w_schema["name"].split(".")[-1]
            r_unqual_name = r_schema["name"].split(".")[-1]
            if w_unqual_name == r_unqual_name or w_schema["name"] in r_schema.get(
                "aliases", []
            ):
                return r_schema
        elif w_type not in AVRO_TYPES and r_type in NAMED_TYPES:
            if match_types(w_type, r_schema["name"]):
                return r_schema["name"]
        elif match_types(w_type, r_type):
            return r_schema
        raise SchemaResolutionError(error_msg)


cpdef inline read_null(fo):
    """null is written as zero bytes."""
    return None


cpdef inline skip_null(fo):
    """null is written as zero bytes."""
    pass


cpdef inline read_boolean(fo):
    """A boolean is written as a single byte whose value is either 0 (false) or
    1 (true).
    """
    cdef unsigned char ch_temp
    cdef bytes bytes_temp = fo.read(1)
    if len(bytes_temp) == 1:
        # technically 0x01 == true and 0x00 == false, but many languages will
        # cast anything other than 0 to True and only 0 to False
        ch_temp = bytes_temp[0]
        return ch_temp != 0
    else:
        raise ReadError


cpdef inline skip_boolean(fo):
    """A boolean is written as a single byte whose value is either 0 (false) or
    1 (true).
    """
    fo.read(1)


cpdef long64 read_long(fo) except? -1:
    """int and long values are written using variable-length, zig-zag
    coding."""
    cdef ulong64 b
    cdef ulong64 n
    cdef int32 shift
    cdef bytes c = fo.read(1)

    # We do EOF checking only here, since most reader start here
    if not c:
        raise StopIteration

    b = <unsigned char>(c[0])
    n = b & 0x7F
    shift = 7

    while (b & 0x80) != 0:
        c = fo.read(1)
        b = <unsigned char>(c[0])
        n |= (b & 0x7F) << shift
        shift += 7

    return (n >> 1) ^ -(n & 1)


cpdef skip_long(fo):
    """int and long values are written using variable-length, zig-zag
    coding."""
    cdef ulong64 b
    cdef bytes c = fo.read(1)

    b = <unsigned char>(c[0])

    while (b & 0x80) != 0:
        c = fo.read(1)
        b = <unsigned char>(c[0])


cpdef skip_int(fo):
    skip_long(fo)


cdef union float_uint32:
    float f
    uint32 n


cpdef read_float(fo):
    """A float is written as 4 bytes.

    The float is converted into a 32-bit integer using a method equivalent to
    Java's floatToIntBits and then encoded in little-endian format.
    """
    cdef bytes data
    cdef unsigned char ch_data[4]
    cdef float_uint32 fi
    data = fo.read(4)
    if len(data) == 4:
        ch_data[:4] = data
        fi.n = (ch_data[0]
                | (ch_data[1] << 8)
                | (ch_data[2] << 16)
                | (ch_data[3] << 24))
        return fi.f
    else:
        raise ReadError


cpdef skip_float(fo):
    """A float is written as 4 bytes.

    The float is converted into a 32-bit integer using a method equivalent to
    Java's floatToIntBits and then encoded in little-endian format.
    """
    fo.read(4)


cdef union double_ulong64:
    double d
    ulong64 n


cpdef read_double(fo):
    """A double is written as 8 bytes.

    The double is converted into a 64-bit integer using a method equivalent to
    Java's doubleToLongBits and then encoded in little-endian format.
    """
    cdef bytes data
    cdef unsigned char ch_data[8]
    cdef double_ulong64 dl
    data = fo.read(8)
    if len(data) == 8:
        ch_data[:8] = data
        dl.n = (ch_data[0]
                | (<ulong64>(ch_data[1]) << 8)
                | (<ulong64>(ch_data[2]) << 16)
                | (<ulong64>(ch_data[3]) << 24)
                | (<ulong64>(ch_data[4]) << 32)
                | (<ulong64>(ch_data[5]) << 40)
                | (<ulong64>(ch_data[6]) << 48)
                | (<ulong64>(ch_data[7]) << 56))
        return dl.d
    else:
        raise ReadError


cpdef skip_double(fo):
    """A double is written as 8 bytes.

    The double is converted into a 64-bit integer using a method equivalent to
    Java's doubleToLongBits and then encoded in little-endian format.
    """
    fo.read(8)


cpdef read_bytes(fo):
    """Bytes are encoded as a long followed by that many bytes of data."""
    cdef long64 size = read_long(fo)
    out =  fo.read(<long>size)
    if len(out) != size:
        raise EOFError(f"Expected {size} bytes, read {len(out)}")
    return out


cpdef skip_bytes(fo):
    """Bytes are encoded as a long followed by that many bytes of data."""
    cdef long64 size = read_long(fo)
    fo.read(<long>size)


cpdef unicode read_utf8(fo):
    """A string is encoded as a long followed by that many bytes of UTF-8
    encoded character data.
    """
    return read_bytes(fo).decode()


cpdef skip_utf8(fo):
    """A string is encoded as a long followed by that many bytes of UTF-8
    encoded character data.
    """
    skip_bytes(fo)


cpdef read_fixed(fo, writer_schema):
    """Fixed instances are encoded using the number of bytes declared in the
    schema."""
    out = fo.read(writer_schema["size"])
    if len(out) != writer_schema["size"]:
        raise EOFError(f"Expected {writer_schema['size']} bytes, read {len(out)}")
    return out


cpdef skip_fixed(fo, writer_schema):
    """Fixed instances are encoded using the number of bytes declared in the
    schema."""
    fo.read(writer_schema["size"])


cpdef read_enum(fo, writer_schema, reader_schema):
    """An enum is encoded by a int, representing the zero-based position of the
    symbol in the schema.
    """
    index = read_long(fo)
    symbol = writer_schema["symbols"][index]
    if reader_schema and symbol not in reader_schema["symbols"]:
        default = reader_schema.get("default")
        if default:
            return default
        else:
            symlist = reader_schema["symbols"]
            msg = f"{symbol} not found in reader symbol list {symlist}"
            raise SchemaResolutionError(msg)
    return symbol


cpdef skip_enum(fo):
    """An enum is encoded by a int, representing the zero-based position of the
    symbol in the schema.
    """
    read_long(fo)


cpdef read_array(
    fo,
    writer_schema,
    named_schemas,
    reader_schema=None,
    return_record_name=False,
    return_record_name_override=False,
):
    """Arrays are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many array
    items.  A block with count zero indicates the end of the array.  Each item
    is encoded per the array's item schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block.  The actual
    count in this case is the absolute value of the count written.
    """
    cdef list read_items
    cdef long64 block_count
    cdef long64 i

    read_items = []

    block_count = read_long(fo)

    while block_count != 0:
        if block_count < 0:
            block_count = -block_count
            # Read block size, unused
            read_long(fo)

        if reader_schema:
            for i in range(block_count):
                read_items.append(_read_data(
                    fo,
                    writer_schema["items"],
                    named_schemas,
                    reader_schema["items"],
                    return_record_name,
                    return_record_name_override,
                ))
        else:
            for i in range(block_count):
                read_items.append(_read_data(
                    fo,
                    writer_schema["items"],
                    named_schemas,
                    None,
                    return_record_name,
                    return_record_name_override,
                ))
        block_count = read_long(fo)

    return read_items


cpdef skip_array(fo, writer_schema, named_schemas):
    """Arrays are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many array
    items.  A block with count zero indicates the end of the array.  Each item
    is encoded per the array's item schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block.  The actual
    count in this case is the absolute value of the count written.
    """
    cdef long64 block_count
    cdef long64 i

    block_count = read_long(fo)

    while block_count != 0:
        if block_count < 0:
            block_size = read_long(fo)
            fo.read(block_size)
        else:
            for i in range(block_count):
                _skip_data(fo, writer_schema["items"], named_schemas)

        block_count = read_long(fo)


cpdef read_map(
    fo,
    writer_schema,
    named_schemas,
    reader_schema=None,
    return_record_name=False,
    return_record_name_override=False,
):
    """Maps are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many key/value
    pairs.  A block with count zero indicates the end of the map.  Each item is
    encoded per the map's value schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block.  The actual
    count in this case is the absolute value of the count written.
    """
    cdef dict read_items
    cdef long64 block_count
    cdef long64 i
    cdef unicode key

    read_items = {}
    block_count = read_long(fo)
    while block_count != 0:
        if block_count < 0:
            block_count = -block_count
            # Read block size, unused
            read_long(fo)

        if reader_schema:
            for i in range(block_count):
                key = read_utf8(fo)
                read_items[key] = _read_data(
                    fo,
                    writer_schema["values"],
                    named_schemas,
                    reader_schema["values"],
                    return_record_name,
                    return_record_name_override,
                )
        else:
            for i in range(block_count):
                key = read_utf8(fo)
                read_items[key] = _read_data(
                    fo,
                    writer_schema["values"],
                    named_schemas,
                    None,
                    return_record_name,
                    return_record_name_override,
                )
        block_count = read_long(fo)

    return read_items


cpdef skip_map(fo, writer_schema, named_schemas):
    """Maps are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many key/value
    pairs.  A block with count zero indicates the end of the map.  Each item is
    encoded per the map's value schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block.  The actual
    count in this case is the absolute value of the count written.
    """
    cdef long64 block_count
    cdef long64 i

    block_count = read_long(fo)
    while block_count != 0:
        if block_count < 0:
            block_size = read_long(fo)
            fo.read(block_size)
        else:
            for i in range(block_count):
                skip_utf8(fo)
                _skip_data(fo, writer_schema["values"], named_schemas)

        block_count = read_long(fo)


cpdef read_union(
    fo,
    writer_schema,
    named_schemas,
    reader_schema=None,
    return_record_name=False,
    return_record_name_override=False,
):
    """A union is encoded by first writing a long value indicating the
    zero-based position within the union of the schema of its value.

    The value is then encoded per the indicated schema within the union.
    """
    # schema resolution
    index = read_long(fo)
    idx_schema = writer_schema[index]

    if reader_schema:
        # Handle case where the reader schema is just a single type (not union)
        if not isinstance(reader_schema, list):
            if match_types(idx_schema, reader_schema):
                return _read_data(
                    fo,
                    idx_schema,
                    named_schemas,
                    reader_schema,
                    return_record_name,
                    return_record_name_override,
                )
        else:
            for schema in reader_schema:
                if match_types(idx_schema, schema):
                    return _read_data(
                        fo,
                        idx_schema,
                        named_schemas,
                        schema,
                        return_record_name,
                        return_record_name_override,
                    )
        msg = f"schema mismatch: {writer_schema} not found in {reader_schema}"
        raise SchemaResolutionError(msg)
    else:
        if return_record_name_override and is_nullable_union(writer_schema):
            return _read_data(
                fo, idx_schema, named_schemas, None, return_record_name, return_record_name_override
            )
        elif return_record_name and extract_record_type(idx_schema) == "record":
            return (
                idx_schema["name"],
                _read_data(
                    fo,
                    idx_schema,
                    named_schemas,
                    None,
                    return_record_name,
                    return_record_name_override,
                )
            )
        elif return_record_name and extract_record_type(idx_schema) not in AVRO_TYPES:
            # idx_schema is a named type
            return (
                named_schemas["writer"][idx_schema]["name"],
                _read_data(
                    fo, idx_schema, named_schemas, None, return_record_name, return_record_name_override
                )
            )

        else:
            return _read_data(
                fo, idx_schema, named_schemas, None, return_record_name, return_record_name_override
            )


cpdef skip_union(fo, writer_schema, named_schemas):
    """A union is encoded by first writing a long value indicating the
    zero-based position within the union of the schema of its value.

    The value is then encoded per the indicated schema within the union.
    """
    # schema resolution
    index = read_long(fo)
    _skip_data(fo, writer_schema[index], named_schemas)


cpdef read_record(
    fo,
    writer_schema,
    named_schemas,
    reader_schema=None,
    return_record_name=False,
    return_record_name_override=False,
):
    """A record is encoded by encoding the values of its fields in the order
    that they are declared. In other words, a record is encoded as just the
    concatenation of the encodings of its fields.  Field values are encoded per
    their schema.

    Schema Resolution:
     * the ordering of fields may be different: fields are matched by name.
     * schemas for fields with the same name in both records are resolved
         recursively.
     * if the writer's record contains a field with a name not present in the
         reader's record, the writer's value for that field is ignored.
     * if the reader's record schema has a field that contains a default value,
         and writer's schema does not have a field with the same name, then the
         reader should use the default value from its field.
     * if the reader's record schema has a field with no default value, and
         writer's schema does not have a field with the same name, then the
         field's value is unset.
    """
    record = {}
    if reader_schema is None:
        for field in writer_schema["fields"]:
            record[field["name"]] = _read_data(
                fo,
                field["type"],
                named_schemas,
                None,
                return_record_name,
                return_record_name_override,
            )
    else:
        readers_field_dict = {}
        aliases_field_dict = {}
        for f in reader_schema["fields"]:
            readers_field_dict[f["name"]] = f
            for alias in f.get("aliases", []):
                aliases_field_dict[alias] = f

        for field in writer_schema["fields"]:
            readers_field = readers_field_dict.get(
                field["name"],
                aliases_field_dict.get(field["name"]),
            )
            if readers_field:
                record[readers_field["name"]] = _read_data(
                    fo,
                    field["type"],
                    named_schemas,
                    readers_field["type"],
                    return_record_name,
                    return_record_name_override,
                )
            else:
                _skip_data(fo, field["type"], named_schemas)

        # fill in default values
        if len(readers_field_dict) > len(record):
            writer_fields = [f["name"] for f in writer_schema["fields"]]
            for f_name, field in readers_field_dict.items():
                if f_name not in writer_fields and f_name not in record:
                    if "default" in field:
                        record[field["name"]] = field["default"]
                    else:
                        msg = f"No default value for {field['name']}"
                        raise SchemaResolutionError(msg)

    return record


cpdef skip_record(fo, writer_schema, named_schemas):
    for field in writer_schema["fields"]:
        _skip_data(fo, field["type"], named_schemas)


cpdef maybe_promote(data, writer_type, reader_type):
    if writer_type == "int":
        # No need to promote to long since they are the same type in Python
        if reader_type == "float" or reader_type == "double":
            return float(data)
    if writer_type == "long":
        if reader_type == "float" or reader_type == "double":
            return float(data)
    if writer_type == "string" and reader_type == "bytes":
        return data.encode()
    if writer_type == "bytes" and reader_type == "string":
        return data.decode()
    return data


cpdef _read_data(
    fo,
    writer_schema,
    named_schemas,
    reader_schema=None,
    return_record_name=False,
    return_record_name_override=False,
):
    """Read data from file object according to schema."""

    record_type = extract_record_type(writer_schema)

    if reader_schema:
        reader_schema = match_schemas(writer_schema, reader_schema)

    try:
        if record_type == "null":
            data = read_null(fo)
        elif record_type == "string":
            data = read_utf8(fo)
        elif record_type == "int" or record_type == "long":
            data = read_long(fo)
        elif record_type == "float":
            data = read_float(fo)
        elif record_type == "double":
            data = read_double(fo)
        elif record_type == "boolean":
            data = read_boolean(fo)
        elif record_type == "bytes":
            data = read_bytes(fo)
        elif record_type == "fixed":
            data = read_fixed(fo, writer_schema)
        elif record_type == "enum":
            data = read_enum(fo, writer_schema, reader_schema)
        elif record_type == "array":
            data = read_array(
                fo,
                writer_schema,
                named_schemas,
                reader_schema,
                return_record_name,
                return_record_name_override,
            )
        elif record_type == "map":
            data = read_map(
                fo,
                writer_schema,
                named_schemas,
                reader_schema,
                return_record_name,
                return_record_name_override,
            )
        elif record_type == "union" or record_type == "error_union":
            data = read_union(
                fo,
                writer_schema,
                named_schemas,
                reader_schema,
                return_record_name,
                return_record_name_override,
            )
        elif record_type == "record" or record_type == "error":
            data = read_record(
                fo,
                writer_schema,
                named_schemas,
                reader_schema,
                return_record_name,
                return_record_name_override,
            )
        else:
            return _read_data(
                fo,
                named_schemas["writer"][record_type],
                named_schemas,
                named_schemas["reader"].get(reader_schema),
                return_record_name,
                return_record_name_override,
            )
    except ReadError:
        raise EOFError(f"cannot read {record_type} from {fo}")

    if "logicalType" in writer_schema:
        logical_type = extract_logical_type(writer_schema)
        fn = LOGICAL_READERS.get(logical_type)
        if fn:
            return fn(data, writer_schema, reader_schema)

    if reader_schema is not None:
        return maybe_promote(
            data,
            record_type,
            extract_record_type(reader_schema)
        )
    else:
        return data


cpdef _skip_data(
    fo,
    writer_schema,
    named_schemas,
):
    record_type = extract_record_type(writer_schema)

    if record_type == "null":
        skip_null(fo)
    elif record_type == "string":
        skip_utf8(fo)
    elif record_type == "int" or record_type == "long":
        skip_long(fo)
    elif record_type == "float":
        skip_float(fo)
    elif record_type == "double":
        skip_double(fo)
    elif record_type == "boolean":
        skip_boolean(fo)
    elif record_type == "bytes":
        skip_bytes(fo)
    elif record_type == "fixed":
        skip_fixed(fo, writer_schema)
    elif record_type == "enum":
        skip_enum(fo)
    elif record_type == "array":
        skip_array(fo, writer_schema, named_schemas)
    elif record_type == "map":
        skip_map(fo, writer_schema, named_schemas)
    elif record_type == "union" or record_type == "error_union":
        skip_union(fo, writer_schema, named_schemas)
    elif record_type == "record" or record_type == "error":
        skip_record(fo, writer_schema, named_schemas)
    else:
        _skip_data(fo, named_schemas["writer"][record_type], named_schemas)


cpdef skip_sync(fo, sync_marker):
    """Skip an expected sync marker, complaining if it doesn't match"""
    if fo.read(SYNC_SIZE) != sync_marker:
        raise ValueError("expected sync marker not found")


cpdef null_read_block(fo):
    """Read block in "null" codec."""
    return BytesIO(read_bytes(fo))


cpdef deflate_read_block(fo):
    """Read block in "deflate" codec."""
    data = read_bytes(fo)
    # -15 is the log of the window size; negative indicates "raw" (no
    # zlib headers) decompression.  See zlib.h.
    return BytesIO(zlib.decompressobj(-15).decompress(data))


cpdef bzip2_read_block(fo):
    """Read block in "bzip2" codec."""
    data = read_bytes(fo)
    return BytesIO(bz2.decompress(data))


cpdef xz_read_block(fo):
    length = read_long(fo)
    data = fo.read(length)
    return BytesIO(lzma.decompress(data))


BLOCK_READERS = {
    "null": null_read_block,
    "deflate": deflate_read_block,
    "bzip2": bzip2_read_block,
    "xz": xz_read_block,
}


cpdef snappy_read_block(fo):
    length = read_long(fo)
    data = fo.read(length - 4)
    fo.read(4)  # CRC
    return BytesIO(snappy.decompress(data))


try:
    import snappy
except ImportError:
    BLOCK_READERS["snappy"] = missing_codec_lib("snappy", "python-snappy")
else:
    BLOCK_READERS["snappy"] = snappy_read_block


cpdef zstandard_read_block(fo):
    length = read_long(fo)
    data = fo.read(length)
    return BytesIO(zstd.ZstdDecompressor().decompressobj().decompress(data))


try:
    import zstandard as zstd
except ImportError:
    BLOCK_READERS["zstandard"] = missing_codec_lib("zstandard", "zstandard")
else:
    BLOCK_READERS["zstandard"] = zstandard_read_block


cpdef lz4_read_block(fo):
    length = read_long(fo)
    data = fo.read(length)
    return BytesIO(lz4.block.decompress(data))


try:
    import lz4.block
except ImportError:
    BLOCK_READERS["lz4"] = missing_codec_lib("lz4", "lz4")
else:
    BLOCK_READERS["lz4"] = lz4_read_block


def _iter_avro_records(
    fo,
    header,
    codec,
    writer_schema,
    named_schemas,
    reader_schema,
    return_record_name=False,
    return_record_name_override=False,
):
    cdef int32 i

    sync_marker = header["sync"]

    read_block = BLOCK_READERS.get(codec)
    if not read_block:
        raise ValueError(f"Unrecognized codec: {codec}")

    block_count = 0
    while True:
        block_count = read_long(fo)
        block_fo = read_block(fo)

        for i in range(block_count):
            yield _read_data(
                block_fo,
                writer_schema,
                named_schemas,
                reader_schema,
                return_record_name,
                return_record_name_override,
            )

        skip_sync(fo, sync_marker)


def _iter_avro_blocks(
    fo,
    header,
    codec,
    writer_schema,
    named_schemas,
    reader_schema,
    return_record_name=False,
    return_record_name_override=False,
):
    sync_marker = header["sync"]

    read_block = BLOCK_READERS.get(codec)
    if not read_block:
        raise ValueError(f"Unrecognized codec: {codec}")

    while True:
        offset = fo.tell()
        try:
            num_block_records = read_long(fo)
        except StopIteration:
            return

        block_bytes = read_block(fo)

        skip_sync(fo, sync_marker)

        size = fo.tell() - offset

        yield Block(
            block_bytes, num_block_records, codec, reader_schema,
            writer_schema, named_schemas, offset, size, return_record_name, return_record_name_override
        )


class Block:
    def __init__(
            self,
            bytes_,
            num_records,
            codec,
            reader_schema,
            writer_schema,
            named_schemas,
            offset,
            size,
            return_record_name=False,
            return_record_name_override=False):
        self.bytes_ = bytes_
        self.num_records = num_records
        self.codec = codec
        self.reader_schema = reader_schema
        self.writer_schema = writer_schema
        self._named_schemas = named_schemas
        self.offset = offset
        self.size = size
        self.return_record_name = return_record_name
        self.return_record_name_override = return_record_name_override

    def __iter__(self):
        for i in range(self.num_records):
            yield _read_data(
                self.bytes_,
                self.writer_schema,
                self._named_schemas,
                self.reader_schema,
                self.return_record_name,
                self.return_record_name_override
            )

    def __str__(self):
        return (
            f"Avro block: {len(self.bytes_)} bytes, {self.num_records} records, "
            + f"codec: {self.codec}, position {self.offset}+{self.size}"
        )


class file_reader:
    def __init__(self, fo, reader_schema=None, return_record_name=False, return_record_name_override=False):
        self.fo = fo
        self.return_record_name = return_record_name
        self.return_record_name_override= return_record_name_override
        try:
            self._header = _read_data(self.fo, HEADER_SCHEMA, {}, None,
                                      return_record_name, return_record_name_override)
        except (StopIteration, EOFError):
            raise ValueError("cannot read header - is it an avro file?")

        # `meta` values are bytes. So, the actual decoding has to be external.
        self.metadata = {
            k: v.decode() for k, v in self._header["meta"].items()
        }

        self._schema = json.loads(self.metadata["avro.schema"])
        self.codec = self.metadata.get("avro.codec", "null")

        self._named_schemas = {"writer": {}, "reader": {}}
        if reader_schema:
            self.reader_schema = parse_schema(
                reader_schema, self._named_schemas["reader"], _write_hint=False
            )
            # Older avro files created before we were more strict about
            # defaults might have been writen with a bad default. Since we re-parse
            # the writer schema here, it will now fail. Therefore, if a user
            # provides a reader schema that passes parsing, we will ignore those
            # default errors
            ignore_default_error = True
        else:
            self.reader_schema = None
            ignore_default_error = False

        self.writer_schema = parse_schema(
            self._schema,
            self._named_schemas["writer"],
            _write_hint=False,
            _force=True,
            _ignore_default_error=ignore_default_error,
        )

        self._elems = None

    @property
    def schema(self):
        import warnings
        warnings.warn(
            "The 'schema' attribute is deprecated. Please use 'writer_schema'",
            DeprecationWarning,
        )
        return self._schema

    def __iter__(self):
        if not self._elems:
            raise NotImplementedError
        return self._elems

    def __next__(self):
        return next(self._elems)


class reader(file_reader):
    def __init__(self, fo, reader_schema=None, return_record_name=False, return_record_name_override=False):
        super().__init__(fo, reader_schema, return_record_name, return_record_name_override)

        self._elems = _iter_avro_records(self.fo,
                                         self._header,
                                         self.codec,
                                         self.writer_schema,
                                         self._named_schemas,
                                         self.reader_schema,
                                         self.return_record_name,
                                         self.return_record_name_override)


class block_reader(file_reader):
    def __init__(self, fo, reader_schema=None, return_record_name=False, return_record_name_override=False):
        super().__init__(fo, reader_schema, return_record_name, return_record_name_override)

        self._elems = _iter_avro_blocks(self.fo,
                                        self._header,
                                        self.codec,
                                        self.writer_schema,
                                        self._named_schemas,
                                        self.reader_schema,
                                        self.return_record_name,
                                        self.return_record_name_override)


cpdef schemaless_reader(fo, writer_schema, reader_schema=None,
                        return_record_name=False,
                        return_record_name_override=False):
    if writer_schema == reader_schema:
        # No need for the reader schema if they are the same
        reader_schema = None

    named_schemas = {"writer": {}, "reader": {}}
    writer_schema = parse_schema(writer_schema, named_schemas["writer"])

    if reader_schema:
        reader_schema = parse_schema(reader_schema, named_schemas["reader"])

    return _read_data(
        fo,
        writer_schema,
        named_schemas,
        reader_schema,
        return_record_name,
        return_record_name_override,
    )


cpdef is_avro(path_or_buffer):
    if isinstance(path_or_buffer, str):
        fp = open(path_or_buffer, "rb")
        close = True
    else:
        fp = path_or_buffer
        close = False

    try:
        header = fp.read(len(MAGIC))
        return header == MAGIC
    finally:
        if close:
            fp.close()



import pytest
import random
import time
from datetime import timezone, datetime, timedelta

SEED = time.time()
random.seed(SEED)


def pytest_report_header(config):
    return f"SEED is {SEED}"


@pytest.fixture(scope="function", autouse=True)
def look_for_cython_error(capfd):
    """If an exception happens in cython and we don't handle it properly it
    will usually print a message to stderr saying that an exception was
    ignored. Here we check for that in every test function."""
    yield
    _, err = capfd.readouterr()
    assert "Exception ignored" not in err


def assert_naive_datetime_equal_to_tz_datetime(naive_datetime, tz_datetime):
    # mktime appears to ignore microseconds, so do this manually
    microseconds = int(time.mktime(naive_datetime.timetuple())) * 1000 * 1000
    microseconds += naive_datetime.microsecond
    aware_datetime = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(
        microseconds=microseconds
    )
    assert aware_datetime == tz_datetime



import io

import fastavro
from fastavro.__main__ import CleanJSONEncoder
import json
import pytest

from decimal import Decimal
from io import BytesIO
from uuid import uuid4
import datetime
import sys
import os
from datetime import timezone, timedelta
import numpy as np

from .conftest import assert_naive_datetime_equal_to_tz_datetime


schema = {
    "fields": [
        {"name": "date", "type": {"type": "int", "logicalType": "date"}},
        {
            "name": "timestamp-millis",
            "type": {"type": "long", "logicalType": "timestamp-millis"},
        },
        {
            "name": "timestamp-micros",
            "type": {"type": "long", "logicalType": "timestamp-micros"},
        },
        {"name": "uuid", "type": {"type": "string", "logicalType": "uuid"}},
        {"name": "time-millis", "type": {"type": "int", "logicalType": "time-millis"}},
        {"name": "time-micros", "type": {"type": "long", "logicalType": "time-micros"}},
    ],
    "namespace": "namespace",
    "name": "name",
    "type": "record",
}


def serialize(schema, data):
    bytes_writer = BytesIO()
    fastavro.schemaless_writer(bytes_writer, schema, data)
    return bytes_writer.getvalue()


def deserialize(schema, binary):
    bytes_writer = BytesIO()
    bytes_writer.write(binary)
    bytes_writer.seek(0)

    res = fastavro.schemaless_reader(bytes_writer, schema)
    return res


def test_logical_types():
    data1 = {
        "date": datetime.date.today(),
        "timestamp-millis": datetime.datetime.now(),
        "timestamp-micros": datetime.datetime.now(),
        "uuid": uuid4(),
        "time-millis": datetime.datetime.now().time(),
        "time-micros": datetime.datetime.now().time(),
    }
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert data1["date"] == data2["date"]
    assert_naive_datetime_equal_to_tz_datetime(
        data1["timestamp-micros"],
        data2["timestamp-micros"],
    )
    assert (
        int(data1["timestamp-millis"].microsecond / 1000) * 1000
        == data2["timestamp-millis"].microsecond
    )
    assert (
        int(data1["time-millis"].microsecond / 1000) * 1000
        == data2["time-millis"].microsecond
    )


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info[:2] == (3, 6),
    reason="Python Bug: https://bugs.python.org/issue29097",
)
def test_not_logical_ints():
    data1 = {
        "date": 1,
        "timestamp-millis": 2,
        "timestamp-micros": 3,
        "uuid": uuid4(),
        "time-millis": 4,
        "time-micros": 5,
    }
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    # 1 day since 1970-1-1
    assert data2["date"] == datetime.date(1970, 1, 2)


schema_null = {
    "fields": [
        {"name": "date", "type": ["null", {"type": "int", "logicalType": "date"}]},
    ],
    "namespace": "namespace",
    "name": "name",
    "type": "record",
}


def test_null():
    data1 = {
        # 'date': None,
    }
    binary = serialize(schema_null, data1)
    data2 = deserialize(schema_null, binary)
    assert data2["date"] is None


def test_not_null_datetime():
    data1 = {"date": datetime.datetime(2017, 1, 1)}
    binary = serialize(schema_null, data1)
    data2 = deserialize(schema_null, binary)
    assert data2["date"] == datetime.date(2017, 1, 1)


def test_not_null_date():
    data1 = {"date": datetime.date(2017, 1, 1)}
    binary = serialize(schema_null, data1)
    data2 = deserialize(schema_null, binary)
    assert data2["date"] == datetime.date(2017, 1, 1)


def test_ancient_datetime():
    schema_datetime = {
        "fields": [
            {
                "name": "timestamp-millis",
                "type": {"type": "long", "logicalType": "timestamp-millis"},
            },
            {
                "name": "timestamp-micros",
                "type": {"type": "long", "logicalType": "timestamp-micros"},
            },
        ],
        "namespace": "namespace",
        "name": "name",
        "type": "record",
    }

    data1 = {
        "timestamp-millis": datetime.datetime(1960, 1, 1),
        "timestamp-micros": datetime.datetime(1960, 1, 1),
    }
    binary = serialize(schema_datetime, data1)
    data2 = deserialize(schema_datetime, binary)

    assert_naive_datetime_equal_to_tz_datetime(
        data1["timestamp-millis"], data2["timestamp-millis"]
    )
    assert_naive_datetime_equal_to_tz_datetime(
        data1["timestamp-micros"], data2["timestamp-micros"]
    )


# test bytes decimal
schema_bytes_decimal = {
    "name": "n",
    "namespace": "namespace",
    "type": "bytes",
    "logicalType": "decimal",
    "precision": 20,
    "scale": 3,
}


@pytest.mark.parametrize(
    "input_data, expected_binary",
    [
        (Decimal("0.0"), b"\x02\x00"),
        (Decimal("-0.0"), b"\x02\x00"),
        (Decimal("0.1"), b"\x02d"),
        (Decimal("-0.1"), b"\x02\x9c"),
        (Decimal("0.2"), b"\x04\x00\xc8"),
        (Decimal("-0.2"), b"\x04\xff8"),
        (Decimal("0.456"), b"\x04\x01\xc8"),
        (Decimal("-0.456"), b"\x04\xfe8"),
        (Decimal("2.55"), b"\x04\t\xf6"),
        (Decimal("-2.55"), b"\x04\xf6\n"),
        (Decimal("2.90"), b"\x04\x0bT"),
        (Decimal("-2.90"), b"\x04\xf4\xac"),
        (Decimal("123.456"), b"\x06\x01\xe2@"),
        (Decimal("-123.456"), b"\x06\xfe\x1d\xc0"),
        (Decimal("3245.234"), b"\x061\x84\xb2"),
        (Decimal("-3245.234"), b"\x06\xce{N"),
        (Decimal("9999999999999999.456"), b"\x12\x00\x8a\xc7#\x04\x89\xe7\xfd\xe0"),
        (Decimal("-999999999999999.456"), b"\x10\xf2\x1fILX\x9c\x02 "),
    ],
)
def test_bytes_decimal(input_data, expected_binary):
    binary = serialize(schema_bytes_decimal, input_data)
    assert binary == expected_binary
    output_data = deserialize(schema_bytes_decimal, binary)
    assert input_data == output_data


def test_bytes_decimal_scale():
    data1 = Decimal("123.456678")  # does not fit scale
    with pytest.raises(ValueError):
        serialize(schema_bytes_decimal, data1)


def test_bytes_decimal_precision():
    data1 = Decimal("123456789012345678.901")  # does not fit precision
    with pytest.raises(ValueError):
        serialize(schema_bytes_decimal, data1)


schema_bytes_decimal_leftmost = {
    "name": "n",
    "namespace": "namespace",
    "type": "bytes",
    "logicalType": "decimal",
    "precision": 18,
    "scale": 6,
}


def test_bytes_decimal_leftmost():
    binary = serialize(schema_bytes_decimal_leftmost, b"\xd5F\x80")
    data2 = deserialize(schema_bytes_decimal_leftmost, binary)
    assert Decimal("-2.80") == data2


# test fixed decimal
schema_fixed_decimal = {
    "name": "n",
    "namespace": "namespace",
    "type": "fixed",
    "size": 8,
    "logicalType": "decimal",
    "precision": 15,
    "scale": 3,
}


def test_fixed_decimal_negative():
    data1 = Decimal("-2.90")
    binary = serialize(schema_fixed_decimal, data1)
    data2 = deserialize(schema_fixed_decimal, binary)
    assert data1 == data2
    assert len(binary) == schema_fixed_decimal["size"]


def test_fixed_decimal_zero():
    data1 = Decimal("0.0")
    binary = serialize(schema_fixed_decimal, data1)
    data2 = deserialize(schema_fixed_decimal, binary)
    assert data1 == data2
    assert len(binary) == schema_fixed_decimal["size"]


def test_fixed_decimal_positive():
    data1 = Decimal("123.456")
    binary = serialize(schema_fixed_decimal, data1)
    data2 = deserialize(schema_fixed_decimal, binary)
    assert data1 == data2
    assert len(binary) == schema_fixed_decimal["size"]


def test_fixed_decimal_scale():
    data1 = Decimal("123.456678")  # does not fit scale
    with pytest.raises(ValueError):
        serialize(schema_fixed_decimal, data1)


def test_fixed_decimal_precision():
    data1 = Decimal("123456789012345678.901")  # does not fit precision
    with pytest.raises(ValueError):
        serialize(schema_fixed_decimal, data1)


schema_fixed_decimal_leftmost = {
    "name": "n",
    "namespace": "namespace",
    "type": "fixed",
    "size": 8,
    "logicalType": "decimal",
    "precision": 18,
    "scale": 6,
}


def test_fixed_decimal_binary():
    binary = serialize(schema_fixed_decimal_leftmost, b"\xFF\xFF\xFF\xFF\xFF\xd5F\x80")
    data2 = deserialize(schema_fixed_decimal_leftmost, binary)
    assert Decimal("-2.80") == data2


def test_clean_json_list():
    values = [
        datetime.datetime.now(),
        datetime.date.today(),
        uuid4(),
        Decimal("1.23"),
        bytes(b"\x00\x01\x61\xff"),
    ]
    str_values = [
        values[0].isoformat(),
        values[1].isoformat(),
        str(values[2]),
        str(values[3]),
        values[4].decode("iso-8859-1"),
    ]
    assert json.dumps(values, cls=CleanJSONEncoder) == json.dumps(str_values)


def test_clean_json_dict():
    values = {
        "1": datetime.datetime.now(),
        "2": datetime.date.today(),
        "3": uuid4(),
        "4": Decimal("1.23"),
        "5": bytes(b"\x00\x01\x61\xff"),
    }
    str_values = {
        "1": values["1"].isoformat(),
        "2": values["2"].isoformat(),
        "3": str(values["3"]),
        "4": str(values["4"]),
        "5": values["5"].decode("iso-8859-1"),
    }
    assert json.dumps(values, cls=CleanJSONEncoder) == json.dumps(str_values)


def test_unknown_logical_type():
    unknown_type_schema = {
        "type": "record",
        "name": "t_customers",
        "fields": [
            {
                "name": "name",
                "type": {"type": "string", "logicalType": "varchar", "maxLength": 200},
            },
            {
                "name": "address",
                "type": {
                    "type": "record",
                    "name": "t_address",
                    "fields": [
                        {
                            "name": "street",
                            "type": {
                                "type": "string",
                                "logicalType": "varchar",
                                "maxLength": 240,
                            },
                        },
                        {
                            "name": "city",
                            "type": {
                                "type": "string",
                                "logicalType": "varchar",
                                "maxLength": 80,
                            },
                        },
                        {
                            "name": "zip",
                            "type": {
                                "type": "string",
                                "logicalType": "varchar",
                                "maxLength": 18,
                            },
                        },
                    ],
                },
            },
        ],
    }

    data1 = {
        "name": "foo",
        "address": {"street": "123 street", "city": "city", "zip": "00000"},
    }
    converted = serialize(unknown_type_schema, data1)
    data2 = deserialize(unknown_type_schema, converted)
    assert data1 == data2


def test_default_scale_value():
    schema = {
        "name": "test_default_scale_value",
        "type": "fixed",
        "size": 8,
        "logicalType": "decimal",
        "precision": 15,
    }

    data1 = Decimal("-2")
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert data1 == data2


def test_date_as_string():
    schema = {"name": "test_date_as_string", "type": "int", "logicalType": "date"}

    data1 = "2019-05-06"
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert datetime.date(2019, 5, 6) == data2


@pytest.mark.skipif(
    hasattr(sys, "pypy_version_info"), reason="pandas takes forever to install on pypy"
)
def test_pandas_datetime():
    """https://github.com/gojek/feast/pull/490#issuecomment-590623525"""

    # Import here as pandas is not installed on pypy for testing
    import pandas as pd

    schema = {
        "fields": [
            {
                "name": "timestamp-micros",
                "type": ["null", {"type": "long", "logicalType": "timestamp-micros"}],
            }
        ],
        "name": "test_pandas_datetime",
        "type": "record",
    }

    data1 = {
        "timestamp-micros": pd.to_datetime(
            datetime.datetime.utcnow().replace(tzinfo=timezone.utc)
        )
    }
    assert serialize(schema, data1)


def test_local_timestamp_millis():
    schema = {"type": "long", "logicalType": "local-timestamp-millis"}

    tz_naive = datetime.datetime(1970, 1, 1, 1)
    binary = serialize(schema, tz_naive)
    data2 = deserialize(schema, binary)
    assert tz_naive == data2

    tz_aware = datetime.datetime(1970, 1, 1, 1, tzinfo=timezone(timedelta(hours=5)))
    binary = serialize(schema, tz_aware)
    data2 = deserialize(schema, binary)
    assert tz_naive == data2

    binary = serialize(schema, 3600 * 1000)
    data2 = deserialize(schema, binary)
    assert tz_naive == data2


def test_local_timestamp_micros():
    schema = {"type": "long", "logicalType": "local-timestamp-micros"}

    tz_naive = datetime.datetime(1970, 1, 1, 1)
    binary = serialize(schema, tz_naive)
    data2 = deserialize(schema, binary)
    assert tz_naive == data2

    tz_aware = datetime.datetime(1970, 1, 1, 1, tzinfo=timezone(timedelta(hours=5)))
    binary = serialize(schema, tz_aware)
    data2 = deserialize(schema, binary)
    assert tz_naive == data2

    binary = serialize(schema, 3600 * 1000 * 1000)
    data2 = deserialize(schema, binary)
    assert tz_naive == data2


class Interface:
    def __init__(self, array_interface):
        array_interface["shape"] = tuple(array_interface["shape"])
        self.__array_interface__ = array_interface


def read_ndarray(data, writer_schema, reader_schema):
    return np.array(Interface(data))


def prepare_ndarray(data, schema):
    if hasattr(data, "__array_interface__"):
        array_interface = data.__array_interface__.copy()
        array_interface["data"] = data.tobytes()
        array_interface["shape"] = list(array_interface["shape"])
        return array_interface
    else:
        return data


fastavro.read.LOGICAL_READERS["record-ndarray"] = read_ndarray
fastavro.write.LOGICAL_WRITERS["record-ndarray"] = prepare_ndarray


def test_ndarray():
    schema = {
        "type": "record",
        "name": "ndarray",
        "fields": [
            {"name": "shape", "type": {"type": "array", "items": "int"}},
            {"name": "typestr", "type": "string"},
            {"name": "data", "type": "bytes"},
            {"name": "version", "type": "int"},
        ],
        "logicalType": "ndarray",
    }

    one_d = np.linspace(0, 1, 10)
    binary = serialize(schema, one_d)
    data2 = deserialize(schema, binary)
    np.testing.assert_equal(one_d, data2)

    two_d = np.linspace(0, 1, 10).reshape(2, 5)
    binary = serialize(schema, two_d)
    data2 = deserialize(schema, binary)
    np.testing.assert_equal(two_d, data2)


def test_ndarray_union():
    schema = [
        "float",
        {
            "type": "record",
            "name": "ndarray",
            "fields": [
                {"name": "shape", "type": {"type": "array", "items": "int"}},
                {"name": "typestr", "type": "string"},
                {"name": "data", "type": "bytes"},
                {"name": "version", "type": "int"},
            ],
            "logicalType": "ndarray",
        },
    ]

    one_d = np.linspace(0, 1, 10)
    binary = serialize(schema, one_d)
    data2 = deserialize(schema, binary)
    np.testing.assert_equal(one_d, data2)

    two_d = np.linspace(0, 1, 10).reshape(2, 5)
    binary = serialize(schema, two_d)
    data2 = deserialize(schema, binary)
    np.testing.assert_equal(two_d, data2)


def test_custom_logical_type_json_reader():
    """https://github.com/fastavro/fastavro/issues/597"""

    def decode_custom_json(data, *args, **kwargs):
        return json.loads(data)

    custom_schema = {
        "name": "Issue",
        "type": "record",
        "fields": [
            {
                "name": "issue_json",
                "default": "{}",
                "type": {"type": "string", "logicalType": "custom-json"},
            },
        ],
    }

    fastavro.read.LOGICAL_READERS["string-custom-json"] = decode_custom_json

    custom_json_object = {
        "issue_json": '{"key": "value"}',
    }

    sio = io.StringIO(json.dumps(custom_json_object))
    re1 = fastavro.json_reader(fo=sio, schema=custom_schema)
    assert next(re1) == {"issue_json": {"key": "value"}}



import pytest
from fastavro.schema import to_parsing_canonical_form
from fastavro._schema_common import PRIMITIVES


@pytest.mark.parametrize(
    "original_schema,canonical_form",
    (
        [(primitive, f'"{primitive}"') for primitive in PRIMITIVES]
        + [({"type": primitive}, f'"{primitive}"') for primitive in PRIMITIVES]
    ),
)
def test_primitive_conversion(original_schema, canonical_form):
    assert to_parsing_canonical_form(original_schema) == canonical_form


def test_fullname_conversion():
    schema = {
        "namespace": "namespace",
        "name": "test_fullname_conversion",
        "type": "record",
        "fields": [],
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"namespace.test_fullname_conversion","type":"record","fields":[]}'
    )


def test_fullname_conversion_empty_namespace():
    schema = {
        "namespace": "",
        "name": "test_fullname_conversion_empty_namespace",
        "type": "record",
        "fields": [],
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_fullname_conversion_empty_namespace","type":"record","fields":[]}'
    )


def test_fullname_conversion_no_namespace():
    schema = {
        "name": "test_fullname_conversion_no_namespace",
        "type": "record",
        "fields": [],
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_fullname_conversion_no_namespace","type":"record","fields":[]}'
    )


def test_remove_doc():
    schema = {"name": "test_remove_doc", "type": "record", "fields": [], "doc": "doc"}

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_remove_doc","type":"record","fields":[]}'
    )


def test_remove_aliases():
    schema = {
        "name": "test_remove_aliases",
        "type": "record",
        "fields": [],
        "aliases": "alias",
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_remove_aliases","type":"record","fields":[]}'
    )


def test_record_field_order():
    schema = {
        "fields": [],
        "name": "test_record_field_order",
        "type": "record",
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_record_field_order","type":"record","fields":[]}'
    )


def test_enum_field_order():
    schema = {
        "symbols": ["A", "B"],
        "name": "test_enum_field_order",
        "type": "enum",
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_enum_field_order","type":"enum","symbols":["A","B"]}'
    )


def test_array_field_order():
    schema = {"items": "int", "type": "array"}

    assert to_parsing_canonical_form(schema) == '{"type":"array","items":"int"}'


def test_map_field_order():
    schema = {"values": "int", "type": "map"}

    assert to_parsing_canonical_form(schema) == '{"type":"map","values":"int"}'


def test_fixed_field_order():
    schema = {
        "size": 4,
        "name": "test_fixed_field_order",
        "type": "fixed",
    }

    assert (
        to_parsing_canonical_form(schema)
        == '{"name":"test_fixed_field_order","type":"fixed","size":4}'
    )


@pytest.mark.parametrize(
    "original_schema,canonical_form",
    [
        (
            {
                "type": "array",
                "items": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            '{"type":"array","items":{"name":"Test","type":"enum","symbols":["A","B"]}}',
        ),
        (
            {
                "type": "map",
                "values": {"type": "enum", "name": "Test", "symbols": ["A", "B"]},
            },
            '{"type":"map","values":{"name":"Test","type":"enum","symbols":["A","B"]}}',
        ),
        (
            ["string", "null", "long"],
            '["string","null","long"]',
        ),
        (
            {
                "type": "record",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            '{"name":"Test","type":"record","fields":[{"name":"f","type":"long"}]}',
        ),
        (
            {
                "type": "error",
                "name": "Test",
                "fields": [{"name": "f", "type": "long"}],
            },
            '{"name":"Test","type":"record","fields":[{"name":"f","type":"long"}]}',
        ),
        (
            {
                "type": "record",
                "name": "Node",
                "fields": [
                    {"name": "label", "type": "string"},
                    {"name": "children", "type": {"type": "array", "items": "Node"}},
                ],
            },
            (
                '{"name":"Node","type":"record","fields":[{"name":"label","type'
                + '":"string"},{"name":"children","type":{"type":"array","items'
                + '":"Node"}}]}'
            ),
        ),
        (
            {
                "type": "record",
                "name": "Lisp",
                "fields": [
                    {
                        "name": "value",
                        "type": [
                            "null",
                            "string",
                            {
                                "type": "record",
                                "name": "Cons",
                                "fields": [
                                    {"name": "car", "type": "Lisp"},
                                    {"name": "cdr", "type": "Lisp"},
                                ],
                            },
                        ],
                    },
                ],
            },
            (
                '{"name":"Lisp","type":"record","fields":[{"name":"value","type'
                + '":["null","string",{"name":"Cons","type":"record","fields":['
                + '{"name":"car","type":"Lisp"},{"name":"cdr","type":"Lisp"}]}]'
                + "}]}"
            ),
        ),
        (
            {
                "type": "record",
                "name": "HandshakeRequest",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "clientHash",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {"name": "clientProtocol", "type": ["null", "string"]},
                    {"name": "serverHash", "type": "MD5"},
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            (
                '{"name":"org.apache.avro.ipc.HandshakeRequest","type":"record"'
                + ',"fields":[{"name":"clientHash","type":{"name":"org.apache.a'
                + 'vro.ipc.MD5","type":"fixed","size":16}},{"name":"clientProto'
                + 'col","type":["null","string"]},{"name":"serverHash","type":"'
                + 'org.apache.avro.ipc.MD5"},{"name":"meta","type":["null",{"ty'
                + 'pe":"map","values":"bytes"}]}]}'
            ),
        ),
        (
            {
                "type": "record",
                "name": "HandshakeResponse",
                "namespace": "org.apache.avro.ipc",
                "fields": [
                    {
                        "name": "match",
                        "type": {
                            "type": "enum",
                            "name": "HandshakeMatch",
                            "symbols": ["BOTH", "CLIENT", "NONE"],
                        },
                    },
                    {"name": "serverProtocol", "type": ["null", "string"]},
                    {
                        "name": "serverHash",
                        "type": ["null", {"name": "MD5", "size": 16, "type": "fixed"}],
                    },
                    {
                        "name": "meta",
                        "type": ["null", {"type": "map", "values": "bytes"}],
                    },
                ],
            },
            (
                '{"name":"org.apache.avro.ipc.HandshakeResponse","type":"record'
                + '","fields":[{"name":"match","type":{"name":"org.apache.avro.'
                + 'ipc.HandshakeMatch","type":"enum","symbols":["BOTH","CLIENT"'
                + ',"NONE"]}},{"name":"serverProtocol","type":["null","string"]'
                + '},{"name":"serverHash","type":["null",{"name":"org.apache.av'
                + 'ro.ipc.MD5","type":"fixed","size":16}]},{"name":"meta","type'
                + '":["null",{"type":"map","values":"bytes"}]}]}'
            ),
        ),
        (
            {
                "type": "record",
                "name": "Interop",
                "namespace": "org.apache.avro",
                "fields": [
                    {"name": "intField", "type": "int"},
                    {"name": "longField", "type": "long"},
                    {"name": "stringField", "type": "string"},
                    {"name": "boolField", "type": "boolean"},
                    {"name": "floatField", "type": "float"},
                    {"name": "doubleField", "type": "double"},
                    {"name": "bytesField", "type": "bytes"},
                    {"name": "nullField", "type": "null"},
                    {
                        "name": "arrayField",
                        "type": {"type": "array", "items": "double"},
                    },
                    {
                        "name": "mapField",
                        "type": {
                            "type": "map",
                            "values": {
                                "name": "Foo",
                                "type": "record",
                                "fields": [{"name": "label", "type": "string"}],
                            },
                        },
                    },
                    {
                        "name": "unionField",
                        "type": [
                            "boolean",
                            "double",
                            {"type": "array", "items": "bytes"},
                        ],
                    },
                    {
                        "name": "enumField",
                        "type": {
                            "type": "enum",
                            "name": "Kind",
                            "symbols": ["A", "B", "C"],
                        },
                    },
                    {
                        "name": "fixedField",
                        "type": {"type": "fixed", "name": "MD5", "size": 16},
                    },
                    {
                        "name": "recordField",
                        "type": {
                            "type": "record",
                            "name": "Node",
                            "fields": [
                                {"name": "label", "type": "string"},
                                {
                                    "name": "children",
                                    "type": {"type": "array", "items": "Node"},
                                },
                            ],
                        },
                    },
                ],
            },
            (
                '{"name":"org.apache.avro.Interop","type":"record","fields":[{"'
                + 'name":"intField","type":"int"},{"name":"longField","type":"l'
                + 'ong"},{"name":"stringField","type":"string"},{"name":"boolFi'
                + 'eld","type":"boolean"},{"name":"floatField","type":"float"},'
                + '{"name":"doubleField","type":"double"},{"name":"bytesField",'
                + '"type":"bytes"},{"name":"nullField","type":"null"},{"name":"'
                + 'arrayField","type":{"type":"array","items":"double"}},{"name'
                + '":"mapField","type":{"type":"map","values":{"name":"org.apac'
                + 'he.avro.Foo","type":"record","fields":[{"name":"label","type'
                + '":"string"}]}}},{"name":"unionField","type":["boolean","doub'
                + 'le",{"type":"array","items":"bytes"}]},{"name":"enumField","'
                + 'type":{"name":"org.apache.avro.Kind","type":"enum","symbols"'
                + ':["A","B","C"]}},{"name":"fixedField","type":{"name":"org.ap'
                + 'ache.avro.MD5","type":"fixed","size":16}},{"name":"recordFie'
                + 'ld","type":{"name":"org.apache.avro.Node","type":"record","f'
                + 'ields":[{"name":"label","type":"string"},{"name":"children",'
                + '"type":{"type":"array","items":"org.apache.avro.Node"}}]}}]}'
            ),
        ),
        (
            {
                "type": "record",
                "name": "ipAddr",
                "fields": [
                    {
                        "name": "addr",
                        "type": [
                            {"name": "IPv6", "type": "fixed", "size": 16},
                            {"name": "IPv4", "type": "fixed", "size": 4},
                        ],
                    }
                ],
            },
            (
                '{"name":"ipAddr","type":"record","fields":[{"name":"addr","typ'
                + 'e":[{"name":"IPv6","type":"fixed","size":16},{"name":"IPv4",'
                + '"type":"fixed","size":4}]}]}'
            ),
        ),
        (
            {
                "type": "record",
                "name": "TestDoc",
                "doc": "Doc string",
                "fields": [{"name": "name", "type": "string", "doc": "Doc String"}],
            },
            (
                '{"name":"TestDoc","type":"record","fields":[{"name":"name","ty'
                + 'pe":"string"}]}'
            ),
        ),
        (
            {
                "type": "enum",
                "name": "Test",
                "symbols": ["A", "B"],
                "doc": "Doc String",
            },
            '{"name":"Test","type":"enum","symbols":["A","B"]}',
        ),
    ],
)
def test_random_cases(original_schema, canonical_form):
    # All of these random test cases came from the test cases here:
    # https://github.com/apache/avro/blob/0552c674637dd15b8751ed5181387cdbd81480d5/lang/py3/avro/tests/test_normalization.py
    assert to_parsing_canonical_form(original_schema) == canonical_form



import array
import datetime
from decimal import Decimal
from io import BytesIO
from uuid import uuid4
import fastavro
from .conftest import assert_naive_datetime_equal_to_tz_datetime

schema = {
    "fields": [
        {
            "name": "union_uuid",
            "type": ["null", {"type": "string", "logicalType": "uuid"}],
        },
        {"name": "array_string", "type": {"type": "array", "items": "string"}},
        {
            "name": "multi_union_time",
            "type": [
                "null",
                "string",
                {"type": "long", "logicalType": "timestamp-micros"},
            ],
        },
        {
            "name": "array_bytes_decimal",
            "type": [
                "null",
                {
                    "type": "array",
                    "items": {
                        "type": "bytes",
                        "logicalType": "decimal",
                        "precision": 18,
                        "scale": 6,
                    },
                },
            ],
        },
        {
            "name": "array_fixed_decimal",
            "type": [
                "null",
                {
                    "type": "array",
                    "items": {
                        "type": "fixed",
                        "name": "FixedDecimal",
                        "size": 8,
                        "logicalType": "decimal",
                        "precision": 18,
                        "scale": 6,
                    },
                },
            ],
        },
        {
            "name": "array_record",
            "type": {
                "type": "array",
                "items": {
                    "type": "record",
                    "name": "some_record",
                    "fields": [
                        {"name": "f1", "type": "string"},
                        {
                            "name": "f2",
                            "type": {
                                "type": "bytes",
                                "logicalType": "decimal",
                                "precision": 18,
                                "scale": 6,
                            },
                        },
                    ],
                },
            },
        },
    ],
    "namespace": "namespace",
    "name": "name",
    "type": "record",
}


def serialize(schema, data):
    bytes_writer = BytesIO()
    fastavro.schemaless_writer(bytes_writer, schema, data)
    return bytes_writer.getvalue()


def deserialize(schema, binary):
    bytes_writer = BytesIO()
    bytes_writer.write(binary)
    bytes_writer.seek(0)

    res = fastavro.schemaless_reader(bytes_writer, schema)
    return res


def test_complex_schema():
    data1 = {
        "union_uuid": uuid4(),
        "array_string": ["a", "b", "c"],
        "multi_union_time": datetime.datetime.now(),
        "array_bytes_decimal": [Decimal("123.456")],
        "array_fixed_decimal": [Decimal("123.456")],
        "array_record": [{"f1": "1", "f2": Decimal("123.456")}],
    }
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    assert len(data1) == len(data2)
    for field in [
        "array_string",
        "array_bytes_decimal",
        "array_fixed_decimal",
        "array_record",
    ]:
        assert data1[field] == data2[field]
    assert_naive_datetime_equal_to_tz_datetime(
        data1["multi_union_time"], data2["multi_union_time"]
    )


def test_complex_schema_nulls():
    data1 = {
        "array_string": ["a", "b", "c"],
        "array_record": [{"f1": "1", "f2": Decimal("123.456")}],
    }
    binary = serialize(schema, data1)
    data2 = deserialize(schema, binary)
    data1_compare = data1
    data1_compare.update(
        {
            "multi_union_time": None,
            "array_bytes_decimal": None,
            "array_fixed_decimal": None,
            "union_uuid": None,
        }
    )
    assert data1_compare == data2


def test_array_from_tuple():
    data_list = serialize({"type": "array", "items": "int"}, [1, 2, 3])
    data_tuple = serialize({"type": "array", "items": "int"}, (1, 2, 3))
    assert data_list == data_tuple


def test_array_from_array():
    schema = {"type": "array", "items": "int"}
    using_list = serialize(schema, [1, -2, 3])
    using_array = serialize(schema, array.array("l", [1, -2, 3]))
    assert using_list == using_array
    assert deserialize(schema, using_list) == [1, -2, 3]


def test_bytes_from_bytearray():
    schema = {"type": "bytes"}
    using_bytes = serialize(schema, b"\x00\xf1\x02")
    using_bytearray = serialize(schema, bytearray(b"\x00\xf1\x02"))
    assert using_bytes == using_bytearray
    assert deserialize(schema, using_bytes) == b"\x00\xf1\x02"
